import json
from typing import Any, TYPE_CHECKING

from requests import Response


if TYPE_CHECKING:
    from metasdk import MetaApp


class MediaService:
    def __init__(self, app: "MetaApp") -> None:
        """
        Инициализирует сервис для работы с медиафайлами.

        :param app: Объект MetaApp.
        :return: None.
        """
        self.__app = app
        self.__options: dict[str, Any] = {}

    def persist_one(
        self,
        file_base64_content: str,
        filename: str,
        extension: str,
        mime: str,
        is_private: bool = True,
    ) -> dict[str, Any]:
        """
        Загружает файл в облако.

        :param file_base64_content: Содержимое файла в формате base64.
        :param filename: Имя файла.
        :param extension: Расширение файла.
        :param mime: MIME-тип файла.
        :param is_private: Флаг, указывающий, является ли файл приватным.
        :return: Словарь с ответом от сервера.
        """
        return self.__app.api_call("MediaService", "persist_one", locals(), {})

    def upload(self, file_descriptor: str, settings: dict[str, Any]) -> dict[str, Any]:
        """
        Загружает файл в облако.

        :param file_descriptor: Открытый дескриптор.
        :param settings: Настройки загрузки.
        :return: Словарь с ответом от сервера.
        """
        multipart_form_data = {"file": file_descriptor}
        params = {"settings": json.dumps(settings)}
        dr = self.__app.native_api_call(
            service="media",
            method="upload",
            data=params,
            options=self.__options,
            multipart_form=True,
            multipart_form_data=multipart_form_data,
            stream=False,
            http_path="/api/meta/v1/",
            http_method="POST",
            connect_timeout_sec=60 * 10,
        )
        return json.loads(dr.text)

    def download(self, media_id: str, as_stream: bool = False) -> Response:
        """
        Скачивает указанный файл.

        :param media_id: Идентификатор медиафайла.
        :param as_stream: Флаг необходимости потокового скачивания.
        :return: Ответ от сервера (объект Response библиотеки requests).
        """
        response = self.__app.native_api_call(
            service="media",
            method="d/" + media_id,
            data={},
            options=self.__options,
            multipart_form=False,
            multipart_form_data=None,
            stream=as_stream,
            http_path="/api/meta/v1/",
            http_method="GET",
        )
        return response

    def info(self, media_id: str) -> dict[str, Any]:
        """
        Возвращает информацию по файлу.

        :param media_id: Идентификатор медиафайла.
        :return: Словарь с информацией о файле.
        """
        dr = self.__app.native_api_call(
            service="media",
            method="i/" + media_id,
            data={},
            options=self.__options,
            multipart_form=False,
            multipart_form_data=None,
            stream=False,
            http_path="/api/meta/v1/",
            http_method="GET",
        )
        return json.loads(dr.text)



from datetime import datetime

from metasdk.exceptions import BadParametersError


def check_postgres_datetime_with_tz(value):
    try:
        datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z")
    except ValueError:
        raise BadParametersError(f"Datetime should match format '%Y-%m-%dT%H:%M:%S.%f%z': {value}")


def check_postgres_datetime_without_tz(value):
    try:
        datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        raise BadParametersError(f"Datetime should match format '%Y-%m-%dT%H:%M:%S': {value}")


def check_postgres_date(value):
    value = value.split("T")[0]
    try:
        datetime.strptime(value, "%Y-%m-%d")
    except ValueError:
        raise BadParametersError(f"Date should match format '%Y-%m-%d': {value}")


def check_postgres_period(start, end):
    try:
        start_date = datetime.strptime(start, "%Y-%m-%d")
    except ValueError:
        raise BadParametersError(f"Start date should match format '%Y-%m-%d': {start}")
    try:
        end_date = datetime.strptime(end, "%Y-%m-%d")
    except ValueError:
        raise BadParametersError(f"End date should match format '%Y-%m-%d': {end}")
    if start_date > end_date:
        raise BadParametersError("Start date of period should be less than end date")



import json
import os


class Worker:
    """
    Класс для обработки фоновых задач на запускаторе
    """

    def __init__(self, app, stdin):
        """
        :type app: metasdk.MetaApp
        :type stdin: basestring
        """
        self.__app = app
        self.__raw_tasks = json.loads(stdin)
        self.debug_tasks = None
        self.__current_task = None

    def single_task(self, main_fn=None):
        self.__run(main_fn, 'single')

    def multiple_task(self, main_fn=None):
        self.__run(main_fn, 'multiple')

    def ignore_task(self, main_fn=None):
        self.__run(main_fn, 'ignore')

    def __run(self, main_fn, resolver_type):
        if 'GARPUN_CLOUD_GENERATE' in os.environ:
            # идет генерация чего-то и реально запускать даже тестовую задачу на надо
            return

        tasks = self.__get_tasks()
        if not tasks:
            return

        log = self.__app.log
        try:
            log.set_entity("session_id", tasks[0].get("sessionId"))
            if resolver_type == 'multiple':
                main_fn(tasks)
            elif resolver_type == 'single':
                for task in tasks:
                    self.__current_task = task
                    log.set_entity("task_id", task.get("id"))
                    main_fn(task)
            elif resolver_type == 'ignore':
                for ignore_ in tasks:
                    log.set_entity("task_id", ignore_.get("id"))
                    main_fn()
        except Exception as e:
            log.critical(u'Воркер упал из-за неожиданного исключения', {"e": e})
            os._exit(1)

    def __get_tasks(self):
        tasks = []
        if self.__app.debug:
            tasks = self.debug_tasks

        if not tasks:
            tasks = self.__raw_tasks
        return tasks

    prop = property()

    @prop.getter
    def current_task(self):
        return self.__current_task



import json
import time
from time import sleep
from typing import Callable, Mapping, Any

import requests

from metasdk.exceptions import BadRequestError, UnexpectedError


class StarterService:
    headers = {"Accept": "application/json", "Content-Type": "application/json"}

    def __init__(self, app, db, starter_api_url):
        """
        Прямые запросы к БД скорее всего уйдут в апи запускатора, так как скорее всего пбудет много БД для тасков запускатора, так как
        Если будет 100500 шард, то врядли все будет в одной БД
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}
        self.__data_get_cache = {}
        self.__metadb = db
        self.__starter_api_url = starter_api_url
        self.log = app.log
        self.max_retries = 30

    def update_task_result_data(self, task: Mapping[str, Any], sleep_sec: int | float = 15, **kwargs) -> None:
        """
        Обновляет результат работы таска запускатора.
        :param task: данные таска.
        :param sleep_sec: время ожидания между попытками обновления.
        :param kwargs: дополнительные параметры запроса.
            **requests_timeout: настройка таймаутов request при обращении к Координатору.
        :return: None.
        """
        # импорт тут, так как глобально над классом он не работает
        from metasdk import DEV_STARTER_STUB_URL

        if any([
            not task.get("serviceId"),
            self.__starter_api_url == DEV_STARTER_STUB_URL,
        ]):
            # В этом случае предполагается, что таск запущен локально.
            self.log.info("STARTER DEV. Результат таска условно обновлен", {"task": task})
            return

        self.log.info("Сохраняем состояние в БД", {"result_data": task["result_data"]})
        requests_timeout = kwargs.get("requests_timeout", 15)
        service_id = task.get("serviceId")
        if sleep_sec < 1:
            sleep_sec = 1
        if sleep_sec > 600:
            sleep_sec = 600
        max_tries = 10
        current_try = 0
        while True:
            url = f"{self.__starter_api_url}/services/{service_id}/tasks/updateResultData"
            try:
                resp = requests.post(
                    url=url,
                    data=json.dumps(task),
                    headers=self.headers,
                    timeout=requests_timeout,
                )
                """
                Осуществляем попытки, пока не получим код 200
                Иначе делаем паузу и пробуем снова
                """
                if resp.status_code == 200:
                    return
                else:
                    self.log.warning("Некорректный http статус при обновлении result_data задачи, пробуем снова", {
                        "task": task
                    })
                    current_try = current_try + 1
                    if current_try >= max_tries:
                        self.log.error("Некорректный http статус при обновлении result_data задачи, прерываем выполнение", {
                            "status_code": resp.status_code,
                            "task": task,
                            "response_text": resp.text
                        })
                        raise IOError("Starter response read error: " + resp.text)

            except Exception:
                self.log.warning("Неизвестная ошибка при обновлении result_data задачи, пробуем снова", {
                    "task": task
                })
                current_try = current_try + 1
                if current_try >= max_tries:
                    self.log.error("Неизвестная ошибка при обновлении result_data задачи, прерываем выполнение", {
                        "task": task
                    })
                    raise IOError("Starter response read error")
            time.sleep(sleep_sec)

    def await_task(
            self,
            task_id: str,
            service_id: str,
            callback_fn: Callable[[dict, bool], None] | None = None,
            sleep_sec: int = 15,
            **kwargs,
    ) -> dict[str, Any] | None:
        """
        Подождать выполнения задачи запускатора.

        При работе с Координатором (запросом данных по API) мы будем обращаться к API до тех пор, пока не будет получен корректный ответ

        Пауза делается в начале, а не конце цикла для того, чтобы не сделать запрос слишком рано, когда задача, вероятно, наверняка еще не выполнена
        Тем не менее, при паузе больше 5 секунд она делится на две части - 5 секунд в начале цикла и остальное - после
        Это позволит не ждать слишком долго перед первым запросом и, в то же время, подерживать общую длинну паузы на заданном параметром уровне
        :param task_id: ID задачи, за которой нужно следить.
        :param service_id: ID сервиса.
        :param callback_fn: Функция обратного вызова, в нее будет передаваться task_info и is_finish как признак, что обработка завершена.
        :param sleep_sec: задержка между проверкой по БД. Не рекомендуется делать меньше 10, так как это может очень сильно ударить по производительности БД.
        :param kwargs: дополнительные параметры запроса.
            **requests_timeout: настройка таймаутов request при обращении к Координатору.
        :return: None|dict
        """
        requests_timeout = kwargs.get("requests_timeout", 15)

        sleep_pre_sec = sleep_sec
        sleep_post_sec = 0
        if sleep_sec > 5:
            sleep_pre_sec = 5
            sleep_post_sec = sleep_sec - 5
        max_tries = 10
        current_try = 0
        while True:
            time.sleep(sleep_pre_sec)
            data = {"taskId": task_id}
            url = f"{self.__starter_api_url}/services/{service_id}/tasks/getShortTaskInfo"
            try:
                resp = requests.post(
                    url=url,
                    data=json.dumps(data),
                    headers=self.headers,
                    timeout=requests_timeout,
                )
                """
                Если задача не найдена - возвращать null и правильно его тут обрабатывать!
                """
                if (len(resp.text) > 0):
                    serverResult = json.loads(resp.text)
                    if resp.status_code == 200:
                        self.log.info("Ждем выполнения задачи", {
                            "task_info": serverResult
                        })
                        is_finish = serverResult["status"] != "NEW" and serverResult["status"] != "PROCESSING"
                        if callback_fn:
                            # Уведомляем вызывающего
                            callback_fn(serverResult, is_finish)
                        if is_finish:
                            return serverResult
                else:
                    return None
            except Exception as e:
                self.log.warning("Неизвестная ошибка при выполнении await_task, пробуем снова", {"error": e})
                current_try = current_try + 1
                if current_try >= max_tries:
                    self.log.error("Неизвестная ошибка при выполнении await_task, прерываем выполнение", {"error": e})
                    raise IOError("Starter response read error")
            time.sleep(sleep_post_sec)

    def submit(self, service_id: str, data: dict | None = None, **kwargs) -> dict[str, Any] | None:
        """
        Отправить задачу в запускатор.

        :param service_id: ID службы. Например "meta.docs_generate".
        :param data: Полезная нагрузка задачи.
        :param kwargs: дополнительные параметры запроса.
            **requests_timeout: настройка таймаутов request при обращении к Координатору.
        :return: dict
        """
        requests_timeout = kwargs.get("requests_timeout", 15)

        # импорт тут, так как глобально над классом он не работает
        from metasdk import DEV_STARTER_STUB_URL

        if self.__starter_api_url == DEV_STARTER_STUB_URL:
            self.log.info("STARTER DEV. Задача условно поставлена", {
                "service_id": service_id,
                "data": data,
            })
            return None

        task = {"serviceId": service_id, "data": data}
        url = f"{self.__starter_api_url}/services/{service_id}/tasks"
        last_e = None
        for _idx in range(self.max_retries):
            try:
                resp = requests.post(
                    url=url,
                    data=json.dumps(task),
                    headers=self.headers,
                    timeout=requests_timeout,
                )
                try:
                    return json.loads(resp.text)
                except Exception:
                    raise IOError("Starter response read error: " + resp.text)
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                # При ошибках подключения пытаемся еще раз
                last_e = e
                sleep(3)
        if isinstance(last_e, Exception):
            raise last_e
        return last_e

    def stop_task(self, task_id: str, service_id: str, **kwargs) -> str | None:
        """
        Остановить задачу запускатора.

        :param task_id: ID задачи.
        :param service_id: ID службы. Например "meta.datasource_share".
        :param kwargs: дополнительные параметры запроса.
            **requests_timeout: настройка таймаутов request при обращении к Координатору.
        :return: None
        """
        requests_timeout = kwargs.get("requests_timeout", 15)

        last_e = None
        url = f"{self.__starter_api_url}/services/{service_id}/tasks/{task_id}"
        for _ in range(self.max_retries):
            try:
                resp = requests.delete(url=url, headers=self.headers, timeout=requests_timeout)
                if resp.status_code == 200:
                    return task_id
                elif resp.status_code == 400:
                    error = resp.json()
                    raise BadRequestError("Bad Request", {"error": error})
                else:
                    raise UnexpectedError("Непредвиденная ошибка при остановке задачи", {"error": resp.text})
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                last_e = e
                sleep(3)
        if isinstance(last_e, Exception):
            raise last_e
        return last_e



import json


class CacheService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}

    def warm_page_cache(self, app_id, page_id, entity_id, object_id, params=None):
        if params is None:
            params = {"stateParams": {}}
        response = self.__app.native_api_call('cache', 'warm-page-cache', params, self.__options, get_params={
            "applicationId": app_id,
            "entityPageId": page_id,
            "entityId": entity_id,
            "objectId": object_id,
        })
        return json.loads(response.text)



"""

Супер мелкие функции, которые нужны от 3 использований

"""
import hashlib
import json
import jwt
import re
from itertools import islice


def chunks_generator(iterable, count_items_in_chunk):
    """
    Очень внимательно! Не дает обходить дважды

    :param iterable:
    :param count_items_in_chunk:
    :return:
    """
    iterator = iter(iterable)
    for first in iterator:  # stops when iterator is depleted
        def chunk():  # construct generator for next chunk
            yield first  # yield element from for loop    # noqa: B023
            for more in islice(iterator, count_items_in_chunk - 1):
                yield more  # yield more elements from the iterator

        yield chunk()  # in outer generator, yield next chunk


def chunks(list_, count_items_in_chunk):
    """
    разбить list (l) на куски по n элементов

    :param list_:
    :param count_items_in_chunk:
    :return:
    """
    for i in range(0, len(list_), count_items_in_chunk):
        yield list_[i:i + count_items_in_chunk]


def pretty_json(obj):
    """
    Представить объект в вище json красиво отформатированной строки
    :param obj:
    :return:
    """
    return json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '), ensure_ascii=False)


def get_decode_algorithms(token: str | bytes) -> list[str]:
    """
    Получает алгоритмы декодирования токена.

    jwt.decode() требует передачи списка алгоритмов для декодирования.
    Стандратно это 'HS256' и 'RS256'. На случай другого алгоритма, проверяем заголовок токена.
    :param token: токен.
    :return: список алгоритмов.
    """
    default_algorithms = ["HS256", "RS256"]
    headers = jwt.get_unverified_header(token)
    header_alg = headers.get("alg")
    if header_alg and header_alg not in default_algorithms:
        default_algorithms.append(header_alg)
    return default_algorithms


def decode_jwt(input_text, secure_key):
    """
    Раскодирование строки на основе ключа
    :param input_text: исходная строка
    :param secure_key: секретный ключ
    :return:
    """
    if input_text is None:
        return None

    encoded = (input_text.split(":")[1]).encode('utf-8')
    algorithms = get_decode_algorithms(token=encoded)
    decoded = jwt.decode(encoded, secure_key, algorithms=algorithms)
    return decoded['sub']


def file_hash_sum(file_full_path: str) -> str:
    """
    Хэш от содержимого файла
    :param file_full_path: полный путь к файлу
    :return: Хэш строкой
    """
    with open(file_full_path, "rb") as f:
        h = hashlib.sha256()
        while True:
            data = f.read(8192)
            if not data:
                break
            h.update(data)
    return h.hexdigest()


def split_text_into_chunks(text: str, max_chunk_size: int, normalize_text: bool = True) -> list[str]:
    """
    Разбивает текст на чанки, не разрывая HTML-теги, если это возможно при заданном максимальном размере.

    :param text: Исходный текст с HTML-тегами
    :param max_chunk_size: Максимальный размер чанка (в символах)
    :return: Список чанков.
    """
    if normalize_text:
        text = text.replace("&lt;", "<").replace("&gt;", ">")
        text = text.replace("<br>", "\n").replace("<br/>", "\n")
        text = re.sub(r"<h[1-4]>", "<b>", text)
        text = re.sub(r"</h[1-4]>", "</b>", text)
        text = text.replace("<p>", "").replace("</p>", "")
        text = text.replace("&nbsp;", " ")
        text = text.replace("\n\n", "\n")
        text = text.replace("```html", "")
        text = text.replace("```", "")
        # text = re.sub(r"\n\s*\n", "\n", text)

    # Регулярное выражение для разделения текста на теги и обычный текст
    pattern = re.compile(r"(<[^>]+>)")
    parts = pattern.split(text)

    chunks = []
    current_chunk = ""

    for part in parts:
        part_length = len(part)

        # Проверяем, превышает ли добавление текущей части размер чанка
        if len(current_chunk) + part_length > max_chunk_size:
            if current_chunk:
                chunks.append(current_chunk)
                current_chunk = ""

            # Если часть сама по себе длиннее максимального размера
            if part_length > max_chunk_size:
                # Если это не тег, разбиваем текст на более мелкие части
                if not pattern.match(part):
                    start = 0
                    while start < part_length:
                        end = start + max_chunk_size
                        chunks.append(part[start:end])
                        start = end
                else:
                    # Если это тег, помещаем его в отдельный чанк
                    chunks.append(part)
            else:
                current_chunk = part
        else:
            current_chunk += part

    # Добавляем оставшийся текст в чанки
    if current_chunk:
        chunks.append(current_chunk)

    return chunks



import logging

from metasdk.logger import LOGGER_ENTITY, REQUEST_LOG


def preprocessing(func):
    def wrapper(self, msg, context=None):
        """ Этот декоратор занимается предобработкой входных параметров:
            1. Проверяет context на None.
            2. Добавляет к msg имя класса объекта ошибки. Например: msg + ModuleNotFoundError
         """
        if context is None:
            context = {}

        error_obj = context.get('e')
        if isinstance(error_obj, Exception):
            try:
                msg = msg + ' ' + str(error_obj.__class__.__name__)
            except Exception:
                pass

        return func(self, msg, context)
    return wrapper


async def error_log_middleware(request, handler):
    try:
        Logger.log_request(method=request.method,
                           url=str(request.url),
                           user_agent=request.headers.get("User-Agent", "-"),
                           referrer=request.headers.get("Referrer", "-"),  # TODO check how it work
                           response_status_code=0,
                           remote_ip=request.remote)
    except Exception as e:
        Logger().info("Can't log request", {"request": str(request), "exception": e})
    response = await handler(request)
    try:
        Logger.set_log_request_response_status_code(response.status)
        if response.status >= 400:
            Logger().error("Http request error")
    except Exception as e:
        Logger().info("Can't log request", {"request": str(request), "exception": e, "response": str(response)})
    return response

# Нужно для работы в aiohttp (см. реализацию декоратор middleware в aiohttp)
error_log_middleware.__middleware_version__ = 1


class Logger:
    """
    Прослойка для упрощения апи логгера
    """

    def set_entity(self, key, value):
        if value is None:
            self.remove_entity(key)
        else:
            LOGGER_ENTITY[key] = value

    @staticmethod
    def log_request(method: str, url: str, user_agent: str, response_status_code: int = 0,
                    referrer: str = "-", remote_ip: str = "-"):
        REQUEST_LOG.update({"method": method,
                            "url": url,
                            "userAgent": user_agent,
                            "responseStatusCode": response_status_code,
                            "referrer": referrer,
                            "remoteIp": remote_ip})

    @staticmethod
    def set_log_request_response_status_code(response_status_code: int):
        REQUEST_LOG["responseStatusCode"] = response_status_code

    @staticmethod
    def remove_entity(key):
        LOGGER_ENTITY.pop(key, None)

    @staticmethod
    def info(msg, context=None):
        if context is None:
            context = {}
        logging.info(msg, extra={'context': context})

    @preprocessing
    def warning(self, msg, context):
        logging.warning(msg, extra={'context': context})

    @preprocessing
    def error(self, msg, context):
        logging.error(msg, extra={'context': context})

    @preprocessing
    def critical(self, msg, context):
        logging.critical(msg, extra={'context': context})

    @preprocessing
    def exception(self, msg, context):
        logging.exception(msg, extra={'context': context})
